@@ -12,29 +12,32 @@ one way to generate any given value... but much more importantly, we can do
 most things by construction instead of by filtering.  That's the difference
 between "I'd like it to be faster" and "doesn't finish at all".
 """
-
+import itertools
 import json
 import math
 import re
 from copy import deepcopy
-from json.encoder import _make_iterencode, encode_basestring_ascii  # type: ignore
 from typing import Any, Dict, List, NoReturn, Optional, Tuple, Union
 
 import jsonschema
 from hypothesis.errors import InvalidArgument
 from hypothesis.internal.floats import next_down as ieee_next_down, next_up
 
-# Mypy does not (yet!) support recursive type definitions.
-# (and writing a few steps by hand is a DoS attack on the AST walker in Pytest)
-JSONType = Union[None, bool, float, str, list, Dict[str, Any]]
+from ._encode import JSONType, encode_canonical_json, sort_key
+
 Schema = Dict[str, JSONType]
+JSONSchemaValidator = Union[
+    jsonschema.validators.Draft4Validator,
+    jsonschema.validators.Draft6Validator,
+    jsonschema.validators.Draft7Validator,
+]
 
 # Canonical type strings, in order.
 TYPE_STRINGS = ("null", "boolean", "integer", "number", "string", "array", "object")
 TYPE_SPECIFIC_KEYS = (
     ("number", "multipleOf maximum exclusiveMaximum minimum exclusiveMinimum"),
     ("integer", "multipleOf maximum exclusiveMaximum minimum exclusiveMinimum"),
-    ("string", "maxLength minLength pattern contentEncoding contentMediaType"),
+    ("string", "maxLength minLength pattern format contentEncoding contentMediaType"),
     ("array", "items additionalItems maxItems minItems uniqueItems contains"),
     (
         "object",
@@ -65,69 +68,25 @@ def next_down(val: float) -> float:
     return out
 
 
-def make_validator(
-    schema: Schema,
-) -> Union[
-    jsonschema.validators.Draft3Validator,
-    jsonschema.validators.Draft4Validator,
-    jsonschema.validators.Draft6Validator,
-    jsonschema.validators.Draft7Validator,
-]:
-    validator_cls = jsonschema.validators.validator_for(schema)
-    return validator_cls(schema)
+def _get_validator_class(schema: Schema) -> JSONSchemaValidator:
+    try:
+        validator = jsonschema.validators.validator_for(schema)
+        validator.check_schema(schema)
+    except jsonschema.exceptions.SchemaError:
+        validator = jsonschema.Draft4Validator
+        validator.check_schema(schema)
+    return validator
 
 
-class CanonicalisingJsonEncoder(json.JSONEncoder):
-    def iterencode(self, o: Any, _one_shot: bool = False) -> Any:
-        """Replace a stdlib method, so we encode integer-valued floats as ints."""
-
-        def floatstr(o: float) -> str:
-            # This is the bit we're overriding - integer-valued floats are
-            # encoded as integers, to support JSONschemas's uniqueness.
-            assert math.isfinite(o)
-            if o == int(o):
-                return repr(int(o))
-            return repr(o)
-
-        return _make_iterencode(
-            {},
-            self.default,
-            encode_basestring_ascii,
-            self.indent,
-            floatstr,
-            self.key_separator,
-            self.item_separator,
-            self.sort_keys,
-            self.skipkeys,
-            _one_shot,
-        )(o, 0)
+def make_validator(schema: Schema) -> JSONSchemaValidator:
+    validator = _get_validator_class(schema)
+    return validator(schema)
 
 
 class HypothesisRefResolutionError(jsonschema.exceptions.RefResolutionError):
     pass
 
 
-def encode_canonical_json(value: JSONType) -> str:
-    """Canonical form serialiser, for uniqueness testing."""
-    return json.dumps(value, sort_keys=True, cls=CanonicalisingJsonEncoder)
-
-
-def sort_key(value: JSONType) -> Tuple[int, float, Union[float, str]]:
-    """Return a sort key (type, guess, tiebreak) that can compare any JSON value.
-
-    Sorts scalar types before collections, and within each type tries for a
-    sensible ordering similar to Hypothesis' idea of simplicity.
-    """
-    if value is None:
-        return (0, 0, 0)
-    if isinstance(value, bool):
-        return (1, int(value), 0)
-    if isinstance(value, (int, float)):
-        return (2 if int(value) == value else 3, abs(value), value >= 0)
-    type_key = {str: 4, list: 5, dict: 6}[type(value)]
-    return (type_key, len(value), encode_canonical_json(value))
-
-
 def get_type(schema: Schema) -> List[str]:
     """Return a canonical value for the "type" key.
 
@@ -298,7 +257,7 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         elif isinstance(schema.get(key), (bool, dict)):
             schema[key] = canonicalish(schema[key])
         else:
-            assert key not in schema
+            assert key not in schema, (key, schema[key])
     for key in SCHEMA_OBJECT_KEYS:
         if key in schema:
             schema[key] = {
@@ -461,6 +420,8 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
                     # TODO: else merge schema-dependencies of required properties
                     # into the base schema after adding required back in and being
                     # careful to avoid an infinite loop...
+            if not schema["dependencies"]:
+                schema.pop("dependencies")
         schema["required"] = sorted(reqs)
         max_ = schema.get("maxProperties", float("inf"))
         assert isinstance(max_, (int, float))
@@ -829,8 +790,42 @@ def merged(schemas: List[Any]) -> Optional[Schema]:
                         return None
                     odeps[k] = m
             odeps.update(s.pop("dependencies"))
-
-        # TODO: merge `items` schemas or lists-of-schemas
+        if "items" in out or "items" in s:
+            oitems = out.pop("items", TRUTHY)
+            sitems = s.pop("items", TRUTHY)
+            if isinstance(oitems, list) and isinstance(sitems, list):
+                out["items"] = []
+                out["additionalItems"] = merged(
+                    [
+                        out.get("additionalItems", TRUTHY),
+                        s.get("additionalItems", TRUTHY),
+                    ]
+                )
+                for a, b in itertools.zip_longest(oitems, sitems):
+                    if a is None:
+                        a = out.get("additionalItems", TRUTHY)
+                    elif b is None:
+                        b = s.get("additionalItems", TRUTHY)
+                    out["items"].append(merged([a, b]))
+            elif isinstance(oitems, list):
+                out["items"] = [merged([x, sitems]) for x in oitems]
+                out["additionalItems"] = merged(
+                    [out.get("additionalItems", TRUTHY), sitems]
+                )
+            elif isinstance(sitems, list):
+                out["items"] = [merged([x, oitems]) for x in sitems]
+                out["additionalItems"] = merged(
+                    [s.get("additionalItems", TRUTHY), oitems]
+                )
+            else:
+                out["items"] = merged([oitems, sitems])
+                if out["items"] is None:
+                    return None
+            if isinstance(out["items"], list) and None in out["items"]:
+                return None
+            if out.get("additionalItems", TRUTHY) is None:
+                return None
+            s.pop("additionalItems", None)
 
         # This loop handles the remaining cases.  Notably, we do not attempt to
         # merge distinct values for:
@@ -851,7 +846,7 @@ def merged(schemas: List[Any]) -> Optional[Schema]:
         if out == FALSEY:
             return FALSEY
     assert isinstance(out, dict)
-    jsonschema.validators.validator_for(out).check_schema(out)
+    _get_validator_class(out)
     return out
 
 
